{-# LANGUAGE CPP, BangPatterns, NoMonomorphismRestriction #-}

-- File: allpairs.hs
-- Description: nbody computation: full all-pairs version
-- Author: Prabhat Totoo

-- Compile | Run
-- Seq: $ ghc --make -O2 -fforce-recomp allpairs.hs | allpairs <numbodies> <steps>
-- Par: $ ghc --make -O2 -threaded -rtsopts -fforce-recomp -DPAR -D<EVALSTRAT|PARMONAD|EDEN> allpairs.hs | allpairs <numbodies> <steps> +RTS -Nx

import System.IO
import System.Environment
import System.Time
import System.Random
import Text.Printf
import Data.List

#if defined (PAR)
import Control.DeepSeq
import GHC.Conc (numCapabilities)
#endif

#if defined (EVALSTRAT)
import Control.Parallel
import Control.Parallel.Strategies
#elif defined (PARMONAD)
import Control.Monad.Par
#elif defined (EDEN)
import Control.Parallel.Eden
import Control.Parallel.Eden.EdenSkel.Auxiliary
import Control.Parallel.Eden.EdenSkel.MapSkels
import Control.Parallel.Eden.EdenSkel.WPSkels
import qualified Control.Parallel.Eden.ParPrim as ParPrim
import Control.Parallel.Eden.ParPrim hiding(noPe,selfPe)
#endif

-- a body consists of pos, vel and mass
data Body
	= Body
	{ x	 :: {-# UNPACK #-} !Double	-- pos of x
	, y	 :: {-# UNPACK #-} !Double	-- pos of y
	, z	 :: {-# UNPACK #-} !Double	-- pos of z
	, vx :: {-# UNPACK #-} !Double	-- vel of x
	, vy :: {-# UNPACK #-} !Double	-- vel of y
	, vz :: {-# UNPACK #-} !Double	-- vel of z
	, m :: {-# UNPACK #-} !Double } -- mass
	deriving (Show,Eq)

-- acceleration
data Accel
	= Accel
	{ ax :: {-# UNPACK #-} !Double
	, ay :: {-# UNPACK #-} !Double
	, az :: {-# UNPACK #-} !Double }

#if defined (PAR)
instance NFData Body
instance NFData Accel
#endif

#if defined (EVALSTRAT) || (PARMONAD)
#if defined (ONECHUNKPERCAP)
chunksize xs = (length xs) `quot` (numCapabilities * 1)
#elif defined (TWOCHUNKSPERCAP)
chunksize xs = (length xs) `quot` (numCapabilities * 2)
#elif defined (FOURCHUNKSPERCAP)
chunksize xs = (length xs) `quot` (numCapabilities * 4)
#endif
#endif

#if defined (PARMONAD)
parMapChunk f n xs = concat ( runPar $ parMap (map f) (chunk n xs) )

chunk :: Int -> [a] -> [[a]]
chunk _ [] = []
chunk n xs = as : chunk n bs where (as,bs) = splitAt n xs
#endif

#if defined (EDEN)
 
instance Trans Body
instance Trans Accel

size = 1000
         
parMapFarm f = shuffle . (parMap (map f)) . (unshuffle noPe)

parMapFarmMinus f = shuffle . (parMap (map f)) . (unshuffle numProc)

parMapFarmChunk f xs = concat (parMapFarm (map f) (chunk size xs))

parMapOfflineFarmChunk f xs = concat (map_offlineFarm (map f) (chunk size xs))

workpoolSortedChunk f xs = concat ((workpoolSorted noPe 50) (map f) (chunk size xs))


chunkmap size mapscheme f xs 
  = concat (mapscheme (map f) (block size xs))
  where block k [] = []
        block k xs = ys : block k zs
         where (ys,zs) = splitAt k xs

map_farm_minus = farm (unshuffle numProc) shuffle
map_offlineFarm_minus = offlineFarm (unshuffle numProc) shuffle

numProc = max (noPe-1) 1

#endif

timeStep = 0.001

-- epsilon
eps = 0.01

randomList :: (Random a) => Int -> [a]
randomList seed = randoms (mkStdGen seed)
   
genBody s = Body x y z vx vy vz m
	where [x,y,z,vx,vy,vz,m] = take 7 $ randomList s

-- func to calc time taken between t0 and t1 in sec
secDiff::ClockTime -> ClockTime -> Float
secDiff (TOD secs1 psecs1) (TOD secs2 psecs2) = fromInteger (psecs2 - psecs1) / 1e12 + fromInteger (secs2 - secs1)

main =
	do
		args <- getArgs
		let 
			n = read (args!!0) :: Int -- num of bodies
			steps = read (args!!1) :: Int -- num of steps (iterations)
			bs = [ (genBody i) | i <- [0..n-1] ]
			res1 = foldl' f 0 bs -- do meaningless sum...
		print (res1)
		t1 <- getClockTime
		let bs' = doSteps steps bs
		print $ foldl' f 0 bs'
		t2 <- getClockTime
		putStrLn $ printf "time taken: %.2fs" $ secDiff t1 t2

f acc (Body x y z vx vy vz m) = acc+x+y+z+vx+vy+vz+m

doSteps 0 bs = bs
doSteps s bs = doSteps (s-1) new_bs
	where
#if defined (EVALSTRAT)
#if defined (NOCHUNK)
		new_bs = map (updatePos . updateVel) bs `using` parList rdeepseq  -- no chunking
#else
		new_bs = map (updatePos . updateVel) bs `using` parListChunk (chunksize bs) rdeepseq
#endif
#elif defined (PARMONAD)
#if defined (NOCHUNK)
		new_bs = runPar $ parMap (updatePos . updateVel) bs 
#else
		new_bs = parMapChunk (updatePos . updateVel) (chunksize bs) bs
#endif
#elif defined (EDEN)

#if defined (PARMAP)
		new_bs = parMap f bs
#elif defined (PARMAPFARM)
		new_bs = parMapFarm f bs
#elif defined (PARMAPFARMMINUS)
		new_bs = parMapFarmMinus f bs
#elif defined (PARMAPFARMCHUNK)
		new_bs = parMapFarmChunk f bs
#elif defined (PARMAPOFFLINEFARMCHUNK)
		new_bs = parMapOfflineFarmChunk f bs
#elif defined (WORKPOOLSORTEDCHUNK)
		new_bs = workpoolSortedChunk f bs
#endif

#else
		new_bs = map f bs
#endif
		f = (updatePos . updateVel)
		updatePos (Body x y z vx vy vz m) = Body (x+timeStep*vx) (y+timeStep*vy) (z+timeStep*vz) vx vy vz m

		updateVel b = foldl' deductChange b (map (accel b) bs)

		deductChange (Body x y z vx vy vz m) (Accel ax ay az) = Body x y z (vx-ax) (vy-ay) (vz-az) m

		accel b_i b_j
			| b_i == b_j = Accel 0 0 0
			| otherwise = Accel (dx*jm*mag) (dy*jm*mag) (dz*jm*mag)
				where
					mag = timeStep / (dSquared * distance)
					distance = sqrt (dSquared)
					dSquared = dx*dx + dy*dy + dz*dz + eps
					dx = ix - jx
					dy = iy - jy
					dz = iz - jz
					Body ix iy iz _ _ _ _  = b_i
					Body jx jy jz _ _ _ jm = b_j
